<?php

namespace Teknisa\CTB\Services;

use Teknisa\CTB\Factories\Ctb10000LancamentosContabeisFactory;
use Teknisa\Libs\Util\DateUtil;
use Teknisa\Libs\Util\Environment;
use Teknisa\Libs\Util\NewCode;
use Teknisa\Model\Contabilidade\Entities\Contopectbl;
use Teknisa\Model\Contabilidade\Entities\Lancccusto;
use Teknisa\Model\Contabilidade\Entities\Lanclote;
use Teknisa\Model\Contabilidade\Entities\Lanccont;
use Teknisa\Model\Geral\Entities\Centcustfil;
use Teknisa\Model\Geral\Entities\Contctbl;
use Teknisa\Model\Geral\Entities\Filial;

class Ctb10000LancamentosContabeisService {

    protected $entityManager;
    protected $environment;
    protected $lancamentosFactory;
    protected $newCode;

    public function __construct(\Doctrine\ORM\EntityManager $entityManager, Environment $environment, Ctb10000LancamentosContabeisFactory $lancamentosFactory, NewCode $newCode) {
        $this->entityManager = $entityManager;
        $this->environment = $environment;
        $this->lancamentosFactory = $lancamentosFactory;
        $this->newCode = $newCode;
    }

    /*
     * Função para verificar se uma data está entre outras duas.
     * Necessária para validar se a data do lançamento está dentro do período
     * contábil parametrizado.
     */
    private function betweenDates($compareDate, $beginDate, $endDate) {
        // É necessário truncar a data, removendo a hora, para que isso não
        // influencie na comparação.
        $beginDate = DateUtil::truncateData($beginDate);
        $endDate = DateUtil::truncateData($endDate);
        // Depois de truncada, a data final deve ser incrementada em um dia,
        // para que o último dia parametrizado também possa ser considerado.
        $endDate = DateUtil::adicionaIntervalo($endDate, 1, DateUtil::DIA);

        // Retorna true se a data não for menor que a primeira ou maior que a segunda.
        return !(DateUtil::comparaDatas($compareDate, $beginDate, DateUtil::ANTES) || DateUtil::comparaDatas($compareDate, $endDate, DateUtil::DEPOIS));
    }

    /*
     * Função para verificar se uma data de lançamento é válida.
     */
    public function checkValidDataLancamento($row) {
        $contopectblRepository = $this->entityManager->getRepository(Contopectbl::class);

        // Busca os parâmetros de contabilização da empresa selecionada.
        $contopectbl = $contopectblRepository->findOneBy(array(
            "cdempresa" => $row["CDEMPRESA"],
            "nrorg" => $this->environment->getNrOrgTrab()
        ));

        // Busca os intervalos de datas que permitem novos lançamentos contábeis.
        $dtiniclanc = $contopectbl->getDtiniclanc();
        $dtfinalanc = $contopectbl->getDtfinalanc();
        $dtiniccont = $contopectbl->getDtiniccont();
        $dtfinacont = $contopectbl->getDtfinacont();

        // Acessa a data de lançamento escolhida pelo usuário e cria o objeto do tipo DateTime.
        $row["DTLANCCONT"] = strpos($row["DTLANCCONT"], ":") === false ? $row["DTLANCCONT"] : DateUtil::truncateString($row["DTLANCCONT"]);
        $dtlanccont = DateUtil::getDataDeString($row["DTLANCCONT"], DateUtil::FORMATO_BRASILEIRO, true);

        // Valida a data de acordo com os intervalos.
        if (!$this->betweenDates($dtlanccont, $dtiniclanc, $dtfinalanc)) {
            throw new \Exception("Operação bloqueada! Data do lançamento fora do período contábil.");
        }
        if (!$this->betweenDates($dtlanccont, $dtiniccont, $dtfinacont)) {
            throw new \Exception("Operação bloqueada! Data do lançamento fora do período contábil.");
        }
    }

    /*
     * Função para excluir um lançamento.
     */
    public function deleteLanccont($row) {
        $lanccontRepository = $this->entityManager->getRepository(Lanccont::class);
        $lancloteRepository = $this->entityManager->getRepository(Lanclote::class);
        $lancamentoLote = $lanccontRepository->findOneBy(array(
            "cdempresa" => $row["CDEMPRESA"],
            "nrlanccont" => $row["NRLANCCONT"],
            "idlancinte" => "L",
            "nrorg" => $this->environment->getNrOrgTrab()
        ));
        
        if ($lancamentoLote !== null) {
            $lote = $lancloteRepository->findOneBy(array(
                "cdempresa" => $row["CDEMPRESA"],
                "nrlancdefi" => $row["NRLANCCONT"],
                "nrorg" => $this->environment->getNrOrgTrab()
            ));
            throw new \Exception("Operação bloqueada! Lançamento contábil referente a lote. Favor realizar a reversão do lote para as devidas alterações.");
        }

        $params = array(
            "cdempresa" => $row["CDEMPRESA"],
            "nrlanccont" => $row["NRLANCCONT"],
            "nrorg" => $this->environment->getNrOrgTrab()
        );

        $this->lancamentosFactory->deleteLanccont($params);
    }

    /*
     * Função para retornar o próximo número sequencial de um lançamento.
     */
    private function getNextSequencialLancamento($cdempresa, $nrlanccont, $nrorg) {
        $params = array(
            "CDEMPRESA" => $cdempresa,
            "NRLANCCONT" => $nrlanccont,
            "NRORG" => $nrorg
        );

        $sequencial = $this->lancamentosFactory->getNextSequencialLancamento($params)["NRSEQLANC"];
        return $sequencial === null ? "00001" : substr("0000" . ++$sequencial, -5);
    }

    /*
     * Função para retornar o número de lançamento de acordo com o cabeçalho.
     */
    private function getNrlanccont($cabecalho) {
        if ($cabecalho["lancamentoNovo"]) {
            $nrorg = $this->environment->getNrOrgTrab();
            $cabecalho["DTLANCCONT"] = strpos($cabecalho["DTLANCCONT"], ":") === false ? $cabecalho["DTLANCCONT"] : DateUtil::truncateString($cabecalho["DTLANCCONT"]);
            $dtlanccont = DateUtil::getDataDeString($cabecalho["DTLANCCONT"], DateUtil::FORMATO_BRASILEIRO, true);
            $ano = DateUtil::getAno($dtlanccont);
            $mes = DateUtil::getMes($dtlanccont);
            $dataLancamento = $ano.$mes;
            $codigoSeq = $this->newCode->getCodeByTableName("LANCCONTABIL".$cabecalho["CDEMPRESA"].$dataLancamento, $nrorg, 1, 12);
            $nrlanccont = $dataLancamento . substr($codigoSeq, 6);
            $cabecalho["NRLANCCONT"] = $nrlanccont;
        }
        else {
            $nrlanccont = $cabecalho["NRLANCCONT"];
        }
        return $nrlanccont;
    }

    /*
     * Função para retornar as partidas de um lançamento.
     */
    public function getPartidas($cdempresa, $nrlanccont) {
        // Cria os parâmetros do filtro a ser utilizado na Factory.
        $params = array(
            "CDEMPRESA" => $cdempresa,
            "NRLANCCONT" => $nrlanccont,
            "NRORG" => $this->environment->getNrOrgTrab()
        );
        return $this->lancamentosFactory->getPartidas($params);
    }

    /*
     * Função para retornar os rateios de um lançamento.
     */
    public function getRateios($cdempresa, $nrlanccont) {
        // Cria os parâmetros do filtro a ser utilizado na Factory.
        $params = array(
            "CDEMPRESA" => $cdempresa,
            "NRLANCCONT" => $nrlanccont,
            "NRORG" => $this->environment->getNrOrgTrab()
        );
        return $this->lancamentosFactory->getRateios($params);
    }

    /*
     * Função que valida um lançamento e salva os dados enviados.
     */
    public function saveLanccont($partidas, $cabecalho) {
        $contopectbl = $this->validateCabecalho($cabecalho);
        $this->validatePartidas($partidas, $contopectbl, $cabecalho);

        $contctblRepository = $this->entityManager->getRepository(Contctbl::class);
        $lancccustoRepository = $this->entityManager->getRepository(Lancccusto::class);
        $lanccontRepository = $this->entityManager->getRepository(Lanccont::class);

        $this->entityManager->beginTransaction();
        try {
            $nrorg = $this->environment->getNrOrgTrab();
            $operador = $this->environment->getCdOperador();
            $currentTime = $this->environment->getCurrentTime();

            $utilizaRateios = $contopectbl->getIdcontgere() == 'S';
            $utilizaCodigoComercial = $contopectbl->getIdcontcome() == 'S';

            $cdempresa = $cabecalho["CDEMPRESA"];
            $nrlanccont = $this->getNrlanccont($cabecalho);
            $cabecalho["DTLANCCONT"] = strpos($cabecalho["DTLANCCONT"], ":") === false ? $cabecalho["DTLANCCONT"] : DateUtil::truncateString($cabecalho["DTLANCCONT"]);
            $dtlanccont = DateUtil::getDataDeString($cabecalho["DTLANCCONT"], DateUtil::FORMATO_BRASILEIRO, true);

            if (!(isset($cabecalho["partidasExcluidas"]) && is_array($cabecalho["partidasExcluidas"]))) {
                $cabecalho["partidasExcluidas"] = array();
            }
            foreach($cabecalho["partidasExcluidas"] as $partida) {
                $params = array(
                    "CDEMPRESA" => $cdempresa,
                    "NRLANCCONT" => $nrlanccont,
                    "NRSEQLANC" => $partida["NRSEQLANC"],
                    "NRORG" => $nrorg
                );
                $this->lancamentosFactory->deletePartida($params);
            }
            $this->entityManager->flush();

            foreach($partidas as $partida) {
                if ($partida["partidaNova"] || $partida["partidaAlterada"]) {
                    if ($partida["partidaNova"]) {
                        $nrseqlanc = $this->getNextSequencialLancamento($cdempresa, $nrlanccont, $nrorg);

                        $lanccont = new Lanccont();
                        $lanccont->setCdempresa($cdempresa);
                        $lanccont->setNrlanccont($nrlanccont);
                        $lanccont->setNrseqlanc($nrseqlanc);
                        $lanccont->setDtlanccont($dtlanccont);
                        $lanccont->setCddocucont($cabecalho["CDDOCUCONT"]);
                        $lanccont->setNrorg($nrorg);
                        $lanccont->setDtinclusao($currentTime);
                        $lanccont->setCdoperinclusao($operador);
                        $lanccont->setNrorginclusao($nrorg);
                        $lanccont->setCdoperins($operador);
                    }
                    else {
                        $lanccont = $lanccontRepository->findOneBy(array(
                            "cdempresa" => $cdempresa,
                            "nrlanccont" => $nrlanccont,
                            "nrseqlanc" => $partida["NRSEQLANC"],
                            "nrorg" => $nrorg
                        ));
                        $lanccont->setDtultatu($currentTime);
                        $lanccont->setCdoperultatu($operador);
                        $lanccont->setNrorgultatu($nrorg);
                    }
                    $lanccont->setCdcontctbl($partida["CDCONTCTBL"]);
                    $lanccont->setCdhistpadr($partida["CDHISTPADR"]);
                    $lanccont->setDshistlanc($partida["DSHISTLANC"]);
                    $lanccont->setVrlancament($partida["VRLANCAMENT"]);
                    $lanccont->setIdtipolanc($partida["VRLANCAMENT"] > 0 ? 'C' : 'D');
                    $lanccont->setIdlanccome($utilizaCodigoComercial ? $partida["IDLANCCOME"] : null);
                    $lanccont->setCdcontcome($utilizaCodigoComercial ? $partida["CDCONTCOME"] : null);
                    $lanccont->setDtultatulan($currentTime);
                    $lanccont->setIdlancinte('M');

                    $this->entityManager->persist($lanccont);
                    $this->entityManager->flush();

                    $contctbl = $contctblRepository->findOneBy(array(
                        "cdcontctbl" => $partida["CDCONTCTBL"],
                        "nrorg" => $nrorg
                    ));
                    if ($utilizaRateios && ($contctbl->getIdcontccus() == 'R' || $contctbl->getIdcontccus() == 'G')) {
                        foreach($partida["rateiosExcluidos"] as $rateio) {
                            $lancccusto = $lancccustoRepository->findOneBy(array(
                                "cdempresa" => $cdempresa,
                                "nrlanccont" => $nrlanccont,
                                "nrseqlanc" => $nrseqlanc,
                                "cdfilial" => $rateio["CDFILIAL"],
                                "cdcentcust" => $rateio["CDCENTCUST"],
                                "nrorg" => $nrorg
                            ));
                            $this->entityManager->remove($lancccusto);
                        }
                        $this->entityManager->flush();

                        foreach($partida["rateios"] as $rateio) {
                            if ($rateio["rateioNovo"]) {
                                $lancccusto = new Lancccusto();
                                $lancccusto->setCdempresa($cdempresa);
                                $lancccusto->setNrlanccont($nrlanccont);
                                $lancccusto->setNrseqlanc($lanccont->getNrseqlanc());
                                $lancccusto->setCdfilial($rateio["CDFILIAL"]);
                                $lancccusto->setCdcentcust($rateio["CDCENTCUST"]);
                                $lancccusto->setNrorg($nrorg);
                                $lancccusto->setDtinclusao($currentTime);
                                $lancccusto->setCdoperinclusao($operador);
                                $lancccusto->setNrorginclusao($nrorg);
                            }
                            else {
                                $lancccusto = $lancccustoRepository->findOneBy(array(
                                    "cdempresa" => $cdempresa,
                                    "nrlanccont" => $nrlanccont,
                                    "nrseqlanc" => $lanccont->getNrseqlanc(),
                                    "cdfilial" => $rateio["CDFILIAL"],
                                    "cdcentcust" => $rateio["CDCENTCUST"],
                                    "nrorg" => $nrorg
                                ));
                                $lancccusto->setDtultatu($currentTime);
                                $lancccusto->setCdoperultatu($operador);
                                $lancccusto->setNrorgultatu($nrorg);
                            }
                            $lancccusto->setCdcontctbl($partida["CDCONTCTBL"]);
                            $lancccusto->setVrlanccust($rateio["VRLANCCUST"]);

                            $rateio["DTCOMPCUSTO"] = strpos($rateio["DTCOMPCUSTO"], ":") === false ? $rateio["DTCOMPCUSTO"] : DateUtil::truncateString($rateio["DTCOMPCUSTO"]);
                            $dtcompcusto = DateUtil::getDataDeString($rateio["DTCOMPCUSTO"], DateUtil::FORMATO_BRASILEIRO, true);
                            $lancccusto->setDtcompcusto($dtcompcusto);

                            $this->entityManager->persist($lancccusto);
                            $this->entityManager->flush();
                        }
                    }
                    else {
                        $params = array(
                            "CDEMPRESA" => $cdempresa,
                            "NRLANCCONT" => $nrlanccont,
                            "NRSEQLANC" => $lanccont->getNrseqlanc(),
                            "NRORG" => $nrorg
                        );
                        $this->lancamentosFactory->deleteRateiosPartida($params);
                    }
                }
            }
            $this->entityManager->commit();
            return $nrlanccont;
        } catch(\Exception $e) {
            $this->entityManager->rollback();
            throw $e;
        }
    }

    /*
     * Função que valida os dados gerais de um lançamento.
     */
    private function validateCabecalho($cabecalho) {
        if (empty($cabecalho["CDDOCUCONT"])) {
            throw new \Exception("Documento deve ser informado.");
        }
        if (empty($cabecalho["DTLANCCONT"])) {
            throw new \Exception("Data do lançamento contábil deve ser informada.");
        }
        if (empty($cabecalho["CDEMPRESA"])) {
            throw new \Exception("Empresa deve ser informada.");
        }
        if (empty($cabecalho["NRLANCCONT"]) && empty($cabecalho["lancamentoNovo"])) {
            throw new \Exception("Número do lançamento deve ser informado.");
        }

        $contopectblRepository = $this->entityManager->getRepository(Contopectbl::class);
        $contopectbl = $contopectblRepository->findOneBy(array(
            "cdempresa" => $cabecalho["CDEMPRESA"],
            "nrorg" => $this->environment->getNrOrgTrab()
        ));
        if ($contopectbl === null) {
            throw new \Exception("Não há parâmetro contábil registrado para a empresa selecionda.");
        }


        $cabecalho["DTLANCCONT"] = strpos($cabecalho["DTLANCCONT"], ":") === false ? $cabecalho["DTLANCCONT"] :  DateUtil::truncateString($cabecalho["DTLANCCONT"]);
        $dtlanccont = DateUtil::getDataDeString($cabecalho["DTLANCCONT"], DateUtil::FORMATO_BRASILEIRO, true);
        $dtiniclanc = $contopectbl->getDtiniclanc();
        $dtfinalanc = $contopectbl->getDtfinalanc();
        $dtiniccont = $contopectbl->getDtiniccont();
        $dtfinacont = $contopectbl->getDtfinacont();
        if (!$this->betweenDates($dtlanccont, $dtiniclanc, $dtfinalanc)) {
            throw new \Exception("Operação bloqueada! Data do lançamento fora do período contábil.");
        }
        if (!$this->betweenDates($dtlanccont, $dtiniccont, $dtfinacont)) {
            throw new \Exception("Operação bloqueada! Data do lançamento fora do período contábil.");
        }

        return $contopectbl;
    }

   
    /*
     * Função que valida os dados gerais de um lançamento.
     */
    private function validatePartidas(Array $partidas, Contopectbl $contopectbl, $cabecalho) {
        if (empty($partidas)) {
            throw new \Exception("Operação bloqueada! Nenhuma partida foi informada para o lançamento.");
        }

        $contctblRepository = $this->entityManager->getRepository(Contctbl::class);
        $lancloteRepository = $this->entityManager->getRepository(Lanclote::class);
        $nrorg = $this->environment->getNrOrgTrab();

        $utilizaRateios = $contopectbl->getIdcontgere() == 'S';
        $utilizaCodigoComercial = $contopectbl->getIdcontcome() == 'S';

        $creditos = 0;
        $debitos = 0;
        $valorTotal = 0;

        foreach($partidas as $partida) {
            if (empty($partida["CDCONTREDU"])) {
                throw new \Exception("Código Reduzido deve ser informado.");
            }
            if (empty($partida["CDCONTCTBL"])) {
                throw new \Exception("Conta Contábil deve ser informada.");
            }
            if (empty($partida["CDHISTPADR"])) {
                throw new \Exception("Código do Histórico Padrão deve ser informado.");
            }
            if (empty($partida["DSHISTLANC"])) {
                throw new \Exception("Descrição do Lançamento deve ser informada.");
            }
            if (empty($partida["VRLANCAMENT"])) {
                throw new \Exception("Valor da partida deve ser informado.");
            }
            if ($partida["VRLANCAMENT"] == 0) {
                throw new \Exception("Valor da partida deve ser diferente de zero.");
            }
            if (!empty($partida["IDLANCINTE"]) && $partida["IDLANCINTE"] == 'L') {
                $lote = $lancloteRepository->findOneBy(array(
                    "cdempresa" => $cabecalho["CDEMPRESA"],
                    "nrlancdefi" => $partida["NRLANCCONT"],
                    "nrorg" => $nrorg
                ));
                throw new \Exception("Operação bloqueada! Lançamento contábil referente a lote. Favor realizar a reversão do lote " . $lote->getCdlote() . " para as devidas alterações.");
            }

            $contctbl = $contctblRepository->findOneBy(array(
                "cdcontctbl" => $partida["CDCONTCTBL"],
                "nrorg" => $nrorg
            ));
            if ($contctbl === null) {
                throw new \Exception("Conta contábil inválida.");
            }
            if ($contctbl->getIdtipocont() != "A") {
                throw new \Exception("Conta contábil deve ser do tipo analítica.");
            }
            if ($utilizaCodigoComercial && $contctbl->getIdtipocome() !== null) {
                if (empty($partida["CDCONTCOME"])) {
                    throw new \Exception("Código Comercial deve ser informado.");
                }
                if (empty($partida["IDLANCCOME"])) {
                    throw new \Exception("Identificador do Código Comercial deve ser informado.");
                }
                if ($partida["IDLANCCOME"] != 'B' && $partida["IDLANCCOME"] != 'C' && $partida["IDLANCCOME"] != 'F') {
                    throw new \Exception("Identificador do Código Comercial inválido.");
                }
            }
            if ($utilizaRateios && ($contctbl->getIdcontccus() == 'R' || $contctbl->getIdcontccus() == 'G')) {
                if (empty($partida["rateios"])) {
                    throw new \Exception("Rateio deve ser informado para as contas do tipo Gerencial ou Oficial/Gerencial.");
                }

                $valorRateios = $this->validateRateios($partida["rateios"], $contopectbl, $cabecalho);
                if ($valorRateios != $partida["VRLANCAMENT"]) {
                    throw new \Exception("A soma dos rateios deve ser igual ao valor da partida.");
                }
            }

            if ($partida["VRLANCAMENT"] > 0) {
                $creditos += $partida["VRLANCAMENT"];
            }
            else {
                $debitos += $partida["VRLANCAMENT"];
            }
            $valorTotal += $partida["VRLANCAMENT"];

            // Correção de erros de operações binárias, que podem resultar em
            // valores com 15+ casas decimais.
            $valorTotal = round($valorTotal, 2);
        }

        if ($creditos == 0) {
            throw new \Exception("Partida de crédito deve ser realizada.");
        }
        if ($debitos == 0) {
            throw new \Exception("Partida de débito deve ser realizada.");
        }
        if ($valorTotal != 0) {
            throw new \Exception("Valor total de crédito difere do total de débito.");
        }
    }

    /*
     * Função que valida os rateios de uma partida.
     */
    private function validateRateios(Array $rateios, Contopectbl $contopectbl, $cabecalho) {
        $centcustfilRepository = $this->entityManager->getRepository(Centcustfil::class);
        $filialRepository = $this->entityManager->getRepository(Filial::class);
        $nrorg = $this->environment->getNrOrgTrab();

        $valorTotal = 0;
        $filialCentcust = array();

        foreach($rateios as $rateio) {
            if (empty($rateio["CDFILIAL"])) {
                throw new \Exception("Unidade do rateio deve ser informada.");
            }
            if (empty($rateio["CDCENTCUST"])) {
                throw new \Exception("Centro de Custo do rateio deve ser informado.");
            }
            if (empty($rateio["DTCOMPCUSTO"])) {
                throw new \Exception("Data da Competência do rateio deve ser informado.");
            }

            $filial = $filialRepository->findOneBy(array(
                "cdfilial" => $rateio["CDFILIAL"],
                "nrorg" => $nrorg
            ));
            if ($filial === null) {
                throw new \Exception("Unidade selecionada inválida.");
            }
            if ($filial->getCdempresa() != $cabecalho["CDEMPRESA"]) {
                throw new \Exception("Unidade não relacionada à empresa do lançamento contábil.");
            }

            $cdfilial = $filial->getCdfilial();
            if (!\Zeedhi\Framework\Util\Functions::arrayKeyExists($cdfilial, $filialCentcust)) {
                $filialCentcust[$cdfilial] = array();
            }

            $centcustfil = $centcustfilRepository->findOneBy(array(
                "cdfilial" => $rateio["CDFILIAL"],
                "cdcentcust" => $rateio["CDCENTCUST"],
                "nrorg" => $nrorg
            ));
            if ($centcustfil === null) {
                throw new \Exception("Centro de Custo selecionado inválido.");
            }
            if ($centcustfil->getCdfilial() != $cdfilial) {
                throw new \Exception("Centro de Custo não relacionada à unidade do rateio.");
            }

            $cdcentcust = $centcustfil->getCdcentcust();
            if (in_array($centcustfil->getCdcentcust(), $filialCentcust[$cdfilial])) {
                throw new \Exception("Rateio já informado para a unidade e centro de custo informados.");
            }
            $filialCentcust[$cdfilial][] = $cdcentcust;

            if($rateio["VRLANCCUST"] == 0) {
                throw new \Exception("Valor do rateio deve ser diferente de zero.");
            }
            $valorTotal += $rateio["VRLANCCUST"];
        }

        return $valorTotal;
    }

     /*
     * Função que
     */
    public function rastreioContabil($row) {
        try{
           $rastreio = $this->lancamentosFactory->rastreioContabil(
               $row['0'],
               $row['1']
            );
            if(!empty($rastreio)){
               return $rastreio[0];
            } else {
                return 'Sem Rastreio Contábil';
            }
        } catch (\Exception $e) {
            throw $e;
        }

    }

     public function salvardocumento($cabecalho) {
       //  var_dump($cabecalho);die;
        try{
            //var_dump($cabecalho);die;
           $this->lancamentosFactory->salvardocumento($cabecalho);

        } catch (\Exception $e) {
            throw $e;
        }

    }


}