<?php

namespace Teknisa\CTB\Services;

use Teknisa\CTB\Factories\Ctb13000RecalculoSaldoFiscalFactory;
use Teknisa\Libs\Util\DateUtil;
use Teknisa\Libs\Util\Environment;
use Teknisa\Model\Contabilidade\Entities\Contopectbl;

class Ctb13000RecalculoSaldoFiscalService {

    protected $entityManager;
    protected $environment;
    protected $recalculoFactory;

    public function __construct(\Doctrine\ORM\EntityManager $entityManager, Environment $environment, Ctb13000RecalculoSaldoFiscalFactory $recalculoFactory) {
        $this->entityManager = $entityManager;
        $this->environment = $environment;
        $this->recalculoFactory = $recalculoFactory;
    }

    /*
     * Função para verificar se uma data está entre outras duas.
     * Necessária para validar se a data do lançamento está dentro do período
     * contábil parametrizado.
     */
    private function betweenDates($compareDate, $beginDate, $endDate) {
        // É necessário truncar a data, removendo a hora, para que isso não
        // influencie na comparação.
        $beginDate = DateUtil::truncateData($beginDate);
        $endDate = DateUtil::truncateData($endDate);
        // Depois de truncada, a data final deve ser incrementada em um dia,
        // para que o último dia parametrizado também possa ser considerado.
        $endDate = DateUtil::adicionaIntervalo($endDate, 1, DateUtil::DIA);

        // Retorna true se a data não for menor que a primeira ou maior que a segunda.
        return !(DateUtil::comparaDatas($compareDate, $beginDate, DateUtil::ANTES) || DateUtil::comparaDatas($compareDate, $endDate, DateUtil::DEPOIS));
    }

    /*
     * Função que valida um lançamento e salva os dados enviados.
     */
    public function recalculaSaldoFiscal($row) {
        // Valida a inserção dos dados.
        $this->validateRow($row);

        // Se a data final foi inserida, converte para o último dia do mês.
        if ($row["DTFINAL"] !== null) {
            // Cria o objeto DateTime para a utilização do utilitário.
            $dtfinal = DateUtil::getDataDeString($row["DTFINAL"], DateUtil::FORMATO_BRASILEIRO, true);
            // Busca a String correspondente ao último dia do mês inserido.
            $row["DTFINAL"] = DateUtil::getUltimoDiaDoMes($dtfinal)->format(DateUtil::FORMATO_BRASILEIRO);
        }

        // Cria o objeto DateTime para a utilização do utilitário.
        $dtinicial = DateUtil::getDataDeString($row["DTINICIAL"], DateUtil::FORMATO_BRASILEIRO, true);
        // Busca a String correspondente ao primeiro dia do mês inserido.
        $row["DTINICIAL"] = DateUtil::getPrimeiroDiaDoMes($dtinicial)->format(DateUtil::FORMATO_BRASILEIRO);

        // Valida as datas depois da alteração feita.
        $this->validateDates($row);

        $this->entityManager->beginTransaction();
        try {
            // Separa as variáveis para visualização.
            $cdempresa = $row["CDEMPRESA"];
            $idrecsaldosoc = $row["IDRECSALDOSOC"];
            $cdcontini = $row["CDCONTCTBLINI"];
            $cdcontfin = $row["CDCONTCTBLFIN"];
            $dtinicial = $row["DTINICIAL"];
            $dtfinal = $row["DTFINAL"];

            // Chama a procedure de recálculo.
            $this->recalculoFactory->recalculaSaldoFiscal($cdempresa, $idrecsaldosoc, $cdcontini, $cdcontfin, $dtinicial, $dtfinal);
            $this->entityManager->commit();
        } catch(\Exception $e) {
            $this->entityManager->rollback();
            throw $e;
        }
    }

    /*
     * Função para verificar se a(s) data(s) inserida(s) é(são) válida(s).
     */
    private function validateDates($row) {
        $contopectblRepository = $this->entityManager->getRepository(Contopectbl::class);

        // Busca os parâmetros de contabilização da empresa selecionada.
        $contopectbl = $contopectblRepository->findOneBy(array(
            "cdempresa" => $row["CDEMPRESA"],
            "nrorg" => $this->environment->getNrOrgTrab()
        ));
        if ($contopectbl === null) {
            throw new \Exception("Não há parâmetro contábil registrado para a empresa selecionda.");
        }

        // Busca os intervalos contábeis parametrizados.
        $dtiniclanc = $contopectbl->getDtiniclanc();
        $dtfinalanc = $contopectbl->getDtfinalanc();
        $dtiniccont = $contopectbl->getDtiniccont();
        $dtfinacont = $contopectbl->getDtfinacont();

        // Acessa a data de lançamento escolhida pelo usuário e cria o objeto do tipo DateTime.
        $dtinicial = DateUtil::getDataDeString($row["DTINICIAL"], DateUtil::FORMATO_BRASILEIRO, true);
        // Valida a data de acordo com os intervalos.
        if (!$this->betweenDates($dtinicial, $dtiniclanc, $dtfinalanc)) {
            throw new \Exception("Data Inicial deve estar dentro do período contábil.");
        }
        if (!$this->betweenDates($dtinicial, $dtiniccont, $dtfinacont)) {
            throw new \Exception("Data Inicial deve estar dentro do período contábil.");
        }

        if ($row["DTFINAL"] !== null) {
            // Acessa a data de lançamento escolhida pelo usuário e cria o objeto do tipo DateTime.
            $dtfinal = DateUtil::getDataDeString($row["DTFINAL"], DateUtil::FORMATO_BRASILEIRO, true);
            // Valida a data de acordo com os intervalos.
            if (!$this->betweenDates($dtfinal, $dtiniclanc, $dtfinalanc)) {
                throw new \Exception("Data Final informada fora do período contábil.");
            }
            if (!$this->betweenDates($dtfinal, $dtiniccont, $dtfinacont)) {
                throw new \Exception("Data Final informada fora do período contábil.");
            }
        }

        // Valida se a data inicial é igual ou menor que o primeiro mês depois do último recálculo realizado.
        $params = array(
            "DTREF" => $row["DTINICIAL"],
            "CDEMPRESA" => $row["CDEMPRESA"],
            "NRORG" => $this->environment->getNrOrgTrab()
        );
        $dtultrecalculo = $this->recalculoFactory->validaDataInicial($params);
        
        if($dtultrecalculo){
            $dtultrecalculo = $dtultrecalculo["DATASALDO"];
            // Caso algum dado tenha sido retornado, então a data é inválida.
            if ($dtultrecalculo !== null) {
                throw new \Exception("Existe saldo a ser recalculado antes da data inicial. A última data recalculada pelo sistema foi $dtultrecalculo. Para evitar inconsistências, é preciso rodar o recálculo a partir do mês seguinte.");
            }
        }
    }

    /*
     * Função que valida todos os dados gerais inseridos pelo usuário.
     */
    private function validateRow($row) {
        if (empty($row["CDEMPRESA"])) {
            throw new \Exception("Empresa deve ser informada.");
        }
        if (empty($row["CDCONTCTBLINI"])) {
            throw new \Exception("Conta Inicial deve ser informada.");
        }
        if (empty($row["CDCONTCTBLFIN"])) {
            throw new \Exception("Conta Final deve ser informada.");
        }
        if (empty($row["DTINICIAL"])) {
            throw new \Exception("Data Inicial deve ser informada.");
        }
        if (empty($row["IDRECSALDOSOC"])) {
            throw new \Exception("Recalcular Saldo Societário deve ser informado.");
        }
        if ($row["CDCONTCTBLFIN"] < $row["CDCONTCTBLINI"]) {
            throw new \Exception("Conta inicial deve ser menor ou igual à conta final.");
        }
        if ($row["DTFINAL"] !== null) {
            $dtinicial = DateUtil::getDataDeString($row["DTINICIAL"], DateUtil::FORMATO_BRASILEIRO, true);
            $dtfinal = DateUtil::getDataDeString($row["DTFINAL"], DateUtil::FORMATO_BRASILEIRO, true);
            if (DateUtil::comparaDatas($dtinicial, $dtfinal, DateUtil::DEPOIS)) {
                throw new \Exception("A Data Final, quando informada, deve ser maior ou igual à data inicial informada.");
            }
        }
    }
}